from decimal import Decimal
from unittest import TestCase
from datetime import datetime

from quiffen.core.transactions import Transaction, Split
from quiffen.utils import parse_date


class TestTransaction(TestCase):
    def test_equality(self):
        date_obj = parse_date('01/06/2021')
        trans1 = Transaction(date_obj, Decimal(150))
        trans2 = Transaction(date_obj, Decimal(150))
        self.assertEqual(trans1, trans2)

        date_obj1 = parse_date('01/06/2021')
        date_obj2 = parse_date('01/06/2021', False)
        trans1 = Transaction(date_obj1, Decimal(150))
        trans2 = Transaction(date_obj2, Decimal(150))
        self.assertNotEqual(trans1, trans2)

    def test_str(self):
        date_obj = parse_date('01/06/2021')
        trans = Transaction(date_obj, Decimal(150))
        print(trans)

    def test_repr(self):
        date_obj = parse_date('01/06/2021')
        trans = Transaction(date_obj, Decimal('-150.60'))
        print(repr(trans))

    def test_parse_date(self):
        date_obj = datetime.strptime('01/06/2021', '%d/%m/%Y')
        self.assertEqual(date_obj, parse_date('01/06/2021'))

        date_obj = datetime.strptime('06/01/2021', '%m/%d/%Y')
        self.assertEqual(date_obj, parse_date('06/01/2021', False))

        date_obj = datetime.strptime('Jan 01 21', '%b %d %y')
        self.assertEqual(date_obj, parse_date('Jan 01 21', False))

        date_obj = datetime.strptime('01 Jan 21', '%d %b %y')
        self.assertEqual(date_obj, parse_date('01 Jan 21'))

    def test_from_list(self):
        lst = ['D01/06/2021', 'T-150.60', 'PMe', 'MMemo']
        date_obj = datetime.strptime('01/06/2021', '%d/%m/%Y')
        tr, _, _ = Transaction.from_list(lst)
        self.assertEqual(tr, Transaction(date=date_obj, amount=Decimal('-150.60'), payee='Me', memo='Memo'))

    def test_from_string(self):
        string = 'D01/06/2021\nT-150.60\nPMe\nMMemo'
        date_obj = datetime.strptime('01/06/2021', '%d/%m/%Y')
        tr, _, _ = Transaction.from_string(string)
        self.assertEqual(tr, Transaction(date=date_obj, amount=Decimal('-150.60'), payee='Me', memo='Memo'))

        string = 'D01/06/2021_T-150.60_PMe_MMemo'
        date_obj = datetime.strptime('01/06/2021', '%d/%m/%Y')
        tr, _, _ = Transaction.from_string(string, separator='_')
        self.assertEqual(tr, Transaction(date=date_obj, amount=Decimal('-150.60'), payee='Me', memo='Memo'))

    def test_multiple_categories(self):
        lst = ['D01/06/2021', 'T-150.60', 'PMe', 'MMemo', 'LReptile', 'LLizard']
        date_obj = datetime.strptime('01/06/2021', '%d/%m/%Y')
        _, categories, _ = Transaction.from_list(lst)
        self.assertEqual(categories['Reptile'].children[0].name, 'Lizard')

    def test_add_remove_split(self):
        date_obj = datetime.strptime('01/06/2021', '%d/%m/%Y')
        tr = Transaction(date=date_obj, amount=Decimal(150.0), payee='Me')
        split = Split(amount=Decimal(50))
        self.assertFalse(tr.is_split)

        tr.add_split(split)
        self.assertTrue(tr.is_split)

        tr.remove_split(amount=30)
        self.assertTrue(tr.is_split)

        tr.remove_split(amount=50, date=date_obj)
        self.assertTrue(tr.is_split)

        tr.remove_split(amount=50)
        self.assertFalse(tr.is_split)
    
    def test_to_dict(self):
        date_obj = parse_date('01/06/2021')
        trans = Transaction(date_obj, Decimal(150))
        expected = {'date': date_obj, 'amount': 150}
        self.assertEqual(trans.to_dict(), expected)



from datetime import datetime

from quiffen.core.categories_classes import Category


def parse_date(date_string, day_first=True):
    """Parse a string date of an unknown format and return a datetime object.

    Parameters
    ----------
    date_string : str
        String containing date found in QIF file
    day_first : bool, default=True
        Whether or not the day comes first in the date (e.g. UK date) or after the month (e.g. US date)

    Returns
    -------
    datetime.datetime
        datetime object with the date data from the ``date_string`` parameter.

    Raises
    -------
    ValueError
        If the date cannot be parsed.
    """
    day_first_patterns = ['%d/%m/%Y',
                          '%d-%m-%Y',
                          '%d/%m/%y',
                          '%d-%m-%y',
                          '%d0%B0%Y',  # 0 values instead of spaces for reasons explained below
                          '%d0%B0%y',
                          '%d0%b0%Y',
                          '%d0%b0%y']

    month_first_patterns = ['%m/%d/%Y',
                            '%m-%d-%Y',
                            '%m/%d/%y',
                            '%m-%d-%y',
                            '%B0%d0%Y',
                            '%B0%d0%y',
                            '%b0%d0%Y',
                            '%b0%d0%y']

    if day_first:
        date_patterns = day_first_patterns + month_first_patterns
    else:
        date_patterns = month_first_patterns + day_first_patterns

    # QIF files sometimes use ' ' instead of a 0 or a ' instead of a /
    date_string = date_string.replace(' ', '0')
    date_string = date_string.replace('\'', '/')

    for pattern in date_patterns:
        try:
            return datetime.strptime(date_string, pattern)
        except ValueError:
            pass

    raise ValueError(f'Date string \'{date_string}\' is not in a recognised format.')


def create_categories(new_category, categories):
    """Add ``new_category`` to ``categories`` after first creating necessary hierarchy.

    If ``new_category`` fits under a category already in ``categories``, then it will just be added as a child.

    Parameters
    ----------
    new_category : Category
        The new category to be added to the iterable of categories.
    categories : iterable
        Iterable containing categories. If in a dict, Category objects are the values.

    Returns
    -------
    iterable : categories
        The new iterable, now containing ``new_category``
    """
    # Add categories in hierarchy
    categories_is_dict = isinstance(categories, dict)
    if new_category.hierarchy is not None:
        current_category = new_category
        hierarchy = new_category.hierarchy.split(':')

        for i in range(2, len(hierarchy) + 1):
            # Check if category exists
            category_to_find = hierarchy[-i]
            found = False
            if categories_is_dict:
                for root_category in categories.values():
                    try:
                        parent = root_category.find_category(category_to_find)
                        found = True
                        break
                    except KeyError:
                        pass
            else:
                for root_category in categories:
                    try:
                        parent = root_category.find_category(category_to_find)
                        found = True
                        break
                    except KeyError:
                        pass

            if found:
                if current_category not in parent.children:
                    print(f'Adding {current_category} as child')
                    parent.add_child(current_category)
                break

            parent = Category(category_to_find)
            parent.hierarchy = ':'.join(hierarchy[:-i + 1])
            parent.add_child(current_category)
            current_category = parent

            # If at the top of the hierarchy, add to categories list
            if i == len(hierarchy):
                if categories_is_dict:
                    categories[current_category.name] = current_category
                else:
                    categories.append(current_category)
    else:
        if categories_is_dict:
            categories[new_category.name] = new_category
        else:
            categories.append(new_category)

    return categories



from datetime import datetime
from decimal import Decimal

from quiffen.core.transactions import Transaction, TransactionList
from quiffen.utils import parse_date

VALID_ACCOUNT_TYPES = [
    'Cash',
    'Bank',
    'CCard',
    'Oth A',
    'Oth L',
    'Invoice',
    'Invst'
]


class Account:
    """
    A class representing a QIF account.

    Attributes
    ----------
    transactions : dict
        A dict of header strings as keys and TransactionList objects as values.
    last_header : {None, 'Cash', 'Bank', 'CCard', 'Oth A', 'Oth L', 'Invoice', 'Invst'}
        The last transaction header used. This must be set before transactions can be added.

    Examples
    --------
    Creating an Account instance, then adding a transaction.

    >>> import quiffen
    >>> from datetime import datetime
    >>> acc = quiffen.Account(name='Example name', desc='Some description')
    >>> acc
    Account(name='Example name', desc='Some description')
    >>> tr = quiffen.Transaction(date=datetime.now(), amount=150.0)
    >>> acc.last_header = 'Bank'
    >>> acc.add_transaction(tr)
    >>> acc.transactions
    {'Bank': TransactionList(Transaction(date=datetime.datetime(2021, 7, 2, 18, 31, 47, 817025), amount=150.0))}

    Creating an account instance from a section list in a QIF file.

    >>> import quiffen
    >>> string = '!Account\\nNPersonal Bank Account\\nDMy Personal bank account with Barclays.\\n^\\n'
    >>> acc = quiffen.Account.from_string(string)
    >>> acc
    Account(name='Personal Bank Account', desc='My Personal bank account with Barclays.')
    """

    def __init__(self,
                 name: str,
                 desc: str = None,
                 account_type: str = None,
                 credit_limit: Decimal = None,
                 balance: Decimal = None,
                 date_at_balance: datetime = None
                 ):
        """Initialise an instance of the Account class.

        Parameters
        ----------
        name : str
            The name of the account.
        desc : str, default=None
            A description of the account.
        account_type : {None, 'Cash', 'Bank', 'CCard', 'Oth A', 'Oth L', 'Invoice', 'Invst'}
            The type of account specified in the QIF file. Relates to the type of transactions stored in the account,
            but not on a functional level.
        credit_limit : decimal.Decimal, default=None
            The credit limit on the account.
        balance : decimal.Decimal, default=None
            The balance in the account at `date_at_balance`.
        date_at_balance: datetime.datetime, default=None
            The date at which `balance` was recorded.
        """
        self._name = name
        self._desc = desc

        if account_type and account_type not in VALID_ACCOUNT_TYPES:
            raise ValueError(f'\'{account_type}\' is not a valid account type. Valid account types are '
                             f'{VALID_ACCOUNT_TYPES}')

        self._account_type = account_type
        self._credit_limit = credit_limit
        self._balance = balance
        self._date_at_balance = date_at_balance

        self._transactions = {}
        self._last_header = None

    def __eq__(self, other):
        if not isinstance(other, Account):
            return False
        return self._name == other.name

    def __str__(self):
        properties = ''
        ignore = ['_transactions', '_last_header']
        for (object_property, value) in self.__dict__.items():
            if value and object_property not in ignore:
                properties += f'\n    {object_property.strip("_").replace("_", " ").title()}: {value}'

        return 'Account:' + properties

    def __repr__(self):
        properties = ''
        ignore = ['_transactions', '_last_header']
        for (object_property, value) in self.__dict__.items():
            if value and object_property not in ignore:
                properties += f'{object_property.strip("_")}={repr(value)}, '

        properties = properties.strip(', ')
        return f'Account({properties})'

    @property
    def name(self):
        """"""
        return self._name

    @name.setter
    def name(self, new_name):
        self._name = str(new_name)

    @property
    def desc(self):
        return self._desc

    @desc.setter
    def desc(self, new_desc):
        self._desc = str(new_desc)

    @property
    def account_type(self):
        return self._account_type

    @account_type.setter
    def account_type(self, new_type):
        if new_type not in VALID_ACCOUNT_TYPES:
            raise ValueError('Invalid account type')
        self._account_type = str(new_type)

    @property
    def credit_limit(self):
        return self._credit_limit

    @credit_limit.setter
    def credit_limit(self, new_limit):
        self._credit_limit = Decimal(new_limit)

    @property
    def balance(self):
        return self._balance

    @balance.setter
    def balance(self, new_balance):
        self._balance = Decimal(new_balance)

    @property
    def date_at_balance(self):
        return self._date_at_balance

    @date_at_balance.setter
    def date_at_balance(self, new_date):
        date = parse_date(new_date)
        self._date_at_balance = date

    @property
    def transactions(self):
        return self._transactions

    @transactions.setter
    def transactions(self, new_transactions):
        transactions = TransactionList(*new_transactions)
        self._transactions = transactions

    @property
    def last_header(self):
        return self._last_header

    @last_header.setter
    def last_header(self, new_header):
        if new_header not in VALID_ACCOUNT_TYPES:
            raise ValueError('Invalid header')
        self._last_header = str(new_header)

    @classmethod
    def from_list(cls, lst, day_first=True):
        """Return a class instance from a list of QIF strings.

        Parameters
        ----------
        lst : list of str
            List of strings containing QIF information about the account.
        day_first : bool, default=True
             Whether the day or month comes first in the date.

        Returns
        -------
        Account
            An Account object created from the QIF strings.
        """
        kwargs = {}
        for field in lst:
            field = field.replace('\n', '')

            if not field:
                continue

            line_code = field[0]

            try:
                field_info = field[1:]
            except KeyError:
                field_info = ''

            # Check the QIF line code for banking-related operations, then append to kwargs.
            if line_code == '!':
                continue
            elif line_code == 'N':
                kwargs['name'] = field_info
            elif line_code == 'D':
                kwargs['desc'] = field_info
            elif line_code == 'T':
                kwargs['account_type'] = field_info
            elif line_code == 'L':
                kwargs['credit_limit'] = Decimal( field_info.replace(',', ''))
            elif line_code == '$' or line_code == '£':
                kwargs['balance'] = Decimal( field_info.replace(',', ''))
            elif line_code == '/':
                balance_date = parse_date(field_info, day_first)
                kwargs['date_at_balance'] = balance_date

        return cls(**kwargs)

    @classmethod
    def from_string(cls, string, separator='\n', day_first=True):
        """Return a class instance from a QIF file section string.

        Parameters
        ----------
        string : str
            The string containing the QIF-formatted data.
        separator : str, default='\n'
             The line separator for the QIF file. This probably won't need changing.
        day_first : bool, default=True
             Whether the day or month comes first in the date.

        Returns
        -------
        Account
            An Account object created from the QIF strings.
        """
        property_list = string.split(separator)
        return cls.from_list(property_list, day_first)

    def add_transaction(self, transaction, header=None):
        """Add a transaction to the dict of TransactionList objects.

        Parameters
        ----------
        transaction : Transaction or Investment
            The Transaction-type object to be added.
        header : {None, 'Cash', 'Bank', 'CCard', 'Oth A', 'Oth L', 'Invoice', 'Invst'}
             The header under which the transaction falls. Will be used as a key in the dict of transactions.

        Raises
        -------
        RuntimeError
            If there is no header provided, and no ``last_header`` set.
        """
        if not header and not self._last_header:
            raise RuntimeError('No header provided and no last header present')
        elif not header:
            header = self._last_header
        else:
            header = header.split(':')[-1]
            self._last_header = header

        if header not in self._transactions:
            self._transactions[header] = TransactionList()

        self._transactions[header].append(transaction)

    def to_dict(self, ignore=None, dictify_transactions=True, dictify_splits=True):
        """Return a dict object representing the Account.

        Parameters
        ----------
        ignore : list of str, default=None
             A list of strings of parameters that should be excluded from the dict.
        dictify_transactions : bool, default=True
             Whether the Transaction objects should be converted to dicts or left as they are.
        dictify_splits : bool, default=True
             Whether Splits within Transaction objects should be converted to dicts or left as they are.

        Returns
        -------
        dict
            A dict representing the Account object.
        """
        if ignore is None:
            ignore = []

        res = {key.strip('_'): value for (key, value) in self.__dict__.items()
               if value is not None and key.strip('_') not in ignore}

        if dictify_transactions and 'transactions' not in ignore:
            dicted_transactions = {}
            for (header, tl) in self._transactions.items():
                dict_list = [transaction.to_dict(ignore=ignore, dictify_splits=dictify_splits) for transaction in tl]
                dicted_transactions[header] = dict_list
            res['transactions'] = dicted_transactions

        return res



"""
Quiffen is a Python package for parsing QIF (Quicken Interchange Format) files.

The package allows users to both read QIF files and interact with the contents, and also to create a QIF structure
and then output to either a QIF file, a CSV of transaction data or a pandas DataFrame.
"""
from quiffen.core.qif import Qif
from quiffen.core.transactions import Transaction, Investment, TransactionList, Split
from quiffen.core.accounts import Account
from quiffen.core.categories_classes import Category, Class
